# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tensor Processor APIs."""
import numpy as np

from mindinsight.datavisual.utils.tools import to_int
from mindinsight.utils.exceptions import ParamValueError
from mindinsight.utils.tensor import TensorUtils, MAX_DIMENSIONS_FOR_TENSOR
from mindinsight.conf.constants import MAX_TENSOR_RESPONSE_DATA_SIZE
from mindinsight.datavisual.common.validation import Validation
from mindinsight.datavisual.common.exceptions import StepTensorDataNotInCacheError, TensorNotExistError
from mindinsight.datavisual.common.exceptions import ResponseDataExceedMaxValueError, TensorTooLargeError
from mindinsight.datavisual.data_transform.tensor_container import TensorContainer
from mindinsight.datavisual.processors.base_processor import BaseProcessor
from mindinsight.datavisual.proto_files import mindinsight_anf_ir_pb2 as anf_ir_pb2


class TensorProcessor(BaseProcessor):
    """Tensor Processor."""
    def get_tensors(self, train_ids, tags, step, dims, detail):
        """
        Get tensor data for given train_ids, tags, step, dims and detail.

        Args:
            train_ids (list): Specify list of train job ID.
            tags (list): Specify list of tag.
            step (int): Specify step of tag, it's necessary when detail is equal to 'data'.
            dims (str): Specify dims of step, it's necessary when detail is equal to 'data'.
            detail (str): Specify which data to query, available values: 'stats', 'histogram' and 'data'.

        Returns:
            dict, a dict including the `tensors`.

        """
        Validation.check_param_empty(train_id=train_ids, tag=tags)

        tensors = []
        for train_id in train_ids:
            tensors += self._get_train_tensors(train_id, tags, step, dims, detail)

        return {"tensors": tensors}

    def _get_train_tensors(self, train_id, tags, step, dims, detail):
        """
        Get tensor data for given train_id, tags, step, dims and detail.

        Args:
            train_id (str): Specify list of train job ID.
            tags (list): Specify list of tag.
            step (int): Specify step of tensor, it's necessary when detail is set to 'data'.
            dims (str): Specify dims of tensor, it's necessary when detail is set to 'data'.
            detail (str): Specify which data to query, available values: 'stats', 'histogram' and 'data'.

        Returns:
            list[dict], a list of dictionaries containing the `train_id`, `tag`, `values`.

        Raises:
            TensorNotExistError, If tensor with specific train_id and tag is not exist in cache.
            ParamValueError, If the value of detail is not within available values:
                            'stats', 'histogram' and 'data'.
        """

        tensors_response = []
        for tag in tags:
            try:
                tensors = self._data_manager.list_tensors(train_id, tag)
            except ParamValueError as err:
                raise TensorNotExistError(err.message)

            if tensors and not isinstance(tensors[0].value, TensorContainer):
                raise TensorNotExistError("there is no tensor data in this tag: {}".format(tag))

            if detail is None or detail == 'stats':
                values = self._get_tensors_summary(detail, tensors)
            elif detail == 'data':
                Validation.check_param_empty(step=step, dims=dims)
                # Limit to query max two dimensions for tensor in table view.
                dims = TensorUtils.parse_shape(dims, limit=MAX_DIMENSIONS_FOR_TENSOR)
                step = to_int(step, "step")
                values = self._get_tensors_data(step, dims, tensors)
            elif detail == 'histogram':
                values = self._get_tensors_histogram(tensors)
            else:
                raise ParamValueError('Can not support this value: {} of detail.'.format(detail))

            tensor = {
                "train_id": train_id,
                "tag": tag,
                "values": values
            }
            tensors_response.append(tensor)

        return tensors_response

    def _get_tensors_summary(self, detail, tensors):
        """
        Builds a JSON-serializable object with information about tensor summary.

        Args:
            detail (str): Specify which data to query, detail value is None or 'stats' at this method.
            tensors (list): The list of _Tensor data.

        Returns:
            dict, a dict including the `wall_time`, `step`, and `value' for each tensor.
                    {
                        "wall_time": 0,
                        "step": 0,
                        "value": {
                            "dims": [1],
                            "data_type": "DT_FLOAT32"
                            "statistics": {
                                "max": 0,
                                "min": 0,
                                "avg": 0,
                                "count": 1,
                                "nan_count": 0,
                                "neg_inf_count": 0,
                                "pos_inf_count": 0
                            } This dict is being set when detail is equal to stats.
                        }
                    }
        """
        values = []
        for tensor in tensors:
            # This value is an instance of TensorContainer
            value = tensor.value
            value_dict = {
                "dims": value.dims,
                "data_type": anf_ir_pb2.DataType.Name(value.data_type)
            }
            if detail and detail == 'stats':
                stats = None
                if value.error_code is None:
                    stats = TensorUtils.get_statistics_dict(stats=value.stats, overall_stats=value.stats)
                value_dict.update({"statistics": stats})

            values.append({
                "wall_time": tensor.wall_time,
                "step": tensor.step,
                "value": value_dict
            })

        return values

    def _get_tensors_data(self, step, dims, tensors):
        """
        Builds a JSON-serializable object with information about tensor dims data.

        Args:
            step (int): Specify step of tensor.
            dims (tuple): Specify dims of tensor.
            tensors (list): The list of _Tensor data.

        Returns:
            dict, a dict including the `wall_time`, `step`, and `value' for each tensor.
                    {
                        "wall_time": 0,
                        "step": 0,
                        "value": {
                            "dims": [1],
                            "data_type": "DT_FLOAT32",
                            "data": [[0.1]]
                            "statistics": {
                                "max": 0,
                                "min": 0,
                                "avg": 0,
                                "count": 1,
                                "nan_count": 0,
                                "neg_inf_count": 0,
                                "pos_inf_count": 0
                            }
                        }
                    }

        Raises:
            ResponseDataExceedMaxValueError, If the size of response data exceed max value.
            StepTensorDataNotInCacheError, If query step is not in cache.
        """
        values = []
        step_in_cache = False
        for tensor in tensors:
            # This value is an instance of TensorContainer
            value = tensor.value
            if step != tensor.step:
                continue
            step_in_cache = True
            if value.error_code is not None:
                raise TensorTooLargeError("Step: {}".format(tensor.step))
            res_data = TensorUtils.get_specific_dims_data(value.tensor_value, dims)
            flatten_data = res_data.flatten().tolist()
            if len(flatten_data) > MAX_TENSOR_RESPONSE_DATA_SIZE:
                raise ResponseDataExceedMaxValueError("the size of response data: {} exceed max value: {}."
                                                      .format(len(flatten_data), MAX_TENSOR_RESPONSE_DATA_SIZE))

            def transfer(array):
                if not isinstance(array, np.ndarray):
                    # The list is used here so that len function can be used
                    # when the value of array is `NAN`、｀-INF｀ or ｀INF｀.
                    array = [array]
                transfer_data = [None] * len(array)
                for index, data in enumerate(array):
                    if isinstance(data, np.ndarray):
                        transfer_data[index] = transfer(data)
                    else:
                        if np.isnan(data):
                            transfer_data[index] = 'NAN'
                        elif np.isneginf(data):
                            transfer_data[index] = '-INF'
                        elif np.isposinf(data):
                            transfer_data[index] = 'INF'
                        else:
                            transfer_data[index] = float(data)
                return transfer_data

            stats = TensorUtils.get_statistics_from_tensor(res_data)
            if stats.nan_count + stats.neg_inf_count + stats.pos_inf_count > 0:
                tensor_data = transfer(res_data)
            else:
                tensor_data = res_data.tolist()
            values.append({
                "wall_time": tensor.wall_time,
                "step": tensor.step,
                "value": {
                    "dims": value.dims,
                    "data_type": anf_ir_pb2.DataType.Name(value.data_type),
                    "data": tensor_data,
                    "statistics": TensorUtils.get_statistics_dict(stats=stats, overall_stats=value.stats)
                }
            })
            break
        if not step_in_cache:
            raise StepTensorDataNotInCacheError("this step: {} data may has been dropped.".format(step))

        return values

    def _get_tensors_histogram(self, tensors):
        """
        Builds a JSON-serializable object with information about tensor histogram data.

        Args:
            tensors (list): The list of _Tensor data.

        Returns:
            dict, a dict including the `wall_time`, `step`, and `value' for each tensor.
                    {
                        "wall_time": 0,
                        "step": 0,
                        "value": {
                            "dims": [1],
                            "data_type": "DT_FLOAT32",
                            "histogram_buckets": [[0.1, 0.2, 3]]
                            "statistics": {
                                "max": 0,
                                "min": 0,
                                "avg": 0,
                                "count": 1,
                                "nan_count": 0,
                                "neg_inf_count": 0,
                                "pos_inf_count": 0
                            }
                        }
                    }
        """
        values = []
        for tensor in tensors:
            # This value is an instance of TensorContainer
            value = tensor.value
            if value.error_code is not None:
                raise TensorTooLargeError("Step: {}".format(tensor.step))
            buckets = value.buckets()
            values.append({
                "wall_time": tensor.wall_time,
                "step": tensor.step,
                "value": {
                    "dims": value.dims,
                    "data_type": anf_ir_pb2.DataType.Name(value.data_type),
                    "histogram_buckets": buckets,
                    "statistics": TensorUtils.get_statistics_dict(stats=value.stats, overall_stats=value.stats)
                }
            })

        return values



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Debugger Online server."""
from concurrent import futures

import grpc
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.conf import settings
from mindinsight.debugger.debugger_services.debugger_grpc_server import DebuggerGrpcServer
from mindinsight.debugger.debugger_services.debugger_server_base import DebuggerServerBase
from mindinsight.debugger.proto import debug_grpc_pb2_grpc as grpc_server_base


def get_debugger_hostname():
    """Get hostname for online debugger server."""
    grpc_port = settings.DEBUGGER_PORT if hasattr(settings, 'DEBUGGER_PORT') else 50051
    host = settings.HOST if hasattr(settings, 'HOST') else '[::]'
    hostname = "{}:{}".format(host, grpc_port)
    return hostname


class DebuggerOnlineServer(DebuggerServerBase):
    """Debugger Online Server."""

    def __init__(self, cache_store, context):
        super(DebuggerOnlineServer, self).__init__(cache_store, context)
        self._grpc_server_manager = self.get_grpc_server_manager()

    def run(self):
        self._grpc_server_manager.start()
        log.info("Start grpc server %s", self._context.hostname)
        self._grpc_server_manager.wait_for_termination()

    def get_grpc_server_manager(self):
        """Get grpc server instance according to hostname."""
        if self._context.hostname is None:
            self._context.hostname = get_debugger_hostname()
        grpc_server = DebuggerGrpcServer(self._cache_store)
        grpc_server_manager = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        grpc_server_base.add_EventListenerServicer_to_server(grpc_server, grpc_server_manager)
        grpc_server_manager.add_insecure_port(self._context.hostname)
        return grpc_server_manager

    def stop(self):
        self._grpc_server_manager.stop(grace=None)
        self.join()



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Parse summary file and save it local file."""

import os
import time

from google.protobuf.message import DecodeError

from mindinsight.datavisual.common import exceptions
from mindinsight.datavisual.common.log import parse_summary_logger
from mindinsight.datavisual.proto_files import lazy_read_pb2
from mindinsight.datavisual.data_access.file_handler import FileHandler
from mindinsight.datavisual.data_transform.summary_parser.image_writer import ImageWriter
from mindinsight.datavisual.data_transform.summary_parser.scalar_writer import ScalarWriter

from ..ms_data_loader import _SummaryParser

HEADER_SIZE = 8
CRC_STR_SIZE = 4
MAX_EVENT_STRING = 500000000
SCALAR = 'scalar_value'
IMAGE = 'image'
INFO_INTERVAL = 10
RETRY_TIMES = 2


class EventParser:
    """Parse summary file and save it to local file."""
    def __init__(self, summary_file, output):
        self.summary_file = summary_file
        self._output = output
        self._scalar_writer = ScalarWriter(self._output)
        self._image_writer = ImageWriter(FileHandler.join(self._output, IMAGE))
        self._file_size = 0
        self._process_info = 0
        self._image_check = False
        self._scalar_check = False

    def parse(self):
        """Load summary file and parse file content."""

        summary_file_handler = FileHandler(self.summary_file, 'rb')

        self._file_size = os.path.getsize(self.summary_file)
        # when current parsed size bigger than self._process_info, print process
        self._process_info = self._file_size // INFO_INTERVAL

        parse_summary_logger.info("Loading %s.", self.summary_file)
        result = self._load(summary_file_handler)

        if result:
            warning = ''
            scalar_path = FileHandler.join(self._output, "scalar.csv")
            image_path = FileHandler.join(self._output, IMAGE)

            if not self._image_check:
                warning = warning + " The summary file contains no image."
            else:
                parse_summary_logger.info("Images are written in %s.", image_path)

            if not self._scalar_check:
                warning = warning + " The summary file contains no scalar value."
            else:
                parse_summary_logger.info("Writing scalar data into %s.", scalar_path)

            self._scalar_writer.write()
            if warning:
                parse_summary_logger.warning(warning)

            parse_summary_logger.info("Finished loading %s.", self.summary_file)

    def _load(self, file_handler):
        """
        Load a log file data.

        Args:
            file_handler (FileHandler): A file handler.

        Returns:
            bool, True if the summary file is finished loading.
        """
        crc_check_time = 0
        while True:
            start_offset = file_handler.offset
            try:
                event_str = _SummaryParser.event_load(file_handler)
                if start_offset != file_handler.offset:
                    self._print_process(file_handler)
                crc_check_time = 0
                if event_str is None:
                    return True
                if len(event_str) > MAX_EVENT_STRING:
                    parse_summary_logger.warning("file_path: %s, event string: %d exceeds %d and drop it.",
                                                 file_handler.file_path, len(event_str), MAX_EVENT_STRING)
                    continue
                self._event_parse(event_str)
            except exceptions.CRCLengthFailedError:
                if crc_check_time > RETRY_TIMES:
                    parse_summary_logger.error(
                        "Check crc length failed, please check the summary file integrity, "
                        "the file may be in transfer, file_path: %s, offset=%s.",
                        file_handler.file_path, start_offset)
                    return True
                parse_summary_logger.warning(
                    "Check crc failed, retrying %d/%d times.", crc_check_time + 1, RETRY_TIMES + 1)
                file_handler.reset_offset(start_offset)
                crc_check_time += 1
                time.sleep(0.5)
            except (OSError, DecodeError, exceptions.MindInsightException) as ex:
                parse_summary_logger.error("Parse file fail, detail: %r, file path: %s.", str(ex),
                                           file_handler.file_path)
                return False

    def _print_process(self, file_handler):
        """Prints the current parsing progress based on the progress of the read file."""
        current_offset = file_handler.offset
        if current_offset >= self._process_info:
            parse_summary_logger.info("Current parsing process: %d/%d, %d%%.", current_offset, self._file_size,
                                      100 * current_offset // os.path.getsize(self.summary_file))
            self._process_info += self._file_size // INFO_INTERVAL
            if self._process_info > os.path.getsize(self.summary_file):
                self._process_info = os.path.getsize(self.summary_file)

    def _event_parse(self, event_str):
        """
        Transform `Event` data to event and extract the scalar and image data.

        Args:
            event_str (str): Message event string in summary proto, data read from file handler.
        """

        plugins = [SCALAR, IMAGE]

        event = lazy_read_pb2.Event.FromString(event_str)

        if event.HasField('summary'):
            for value in event.summary.value:
                for plugin in plugins:
                    if not value.HasField(plugin):
                        continue
                    self._parse_summary_value(value.tag, event.step, event.wall_time, value, plugin)

    def _parse_summary_value(self, tag, step, wall_time, value, plugin):
        """
        Parse summary value and write corresponding file according to plugin.

        Args:
            tag (str): value tag
            step (int): train step
            wall_time (float): Timestamp
            value (Summary.Value): Value message in summary file.
            plugin (str): Plugin value.
        """
        if plugin == SCALAR:
            self._scalar_writer.add((tag, step, wall_time, value.scalar_value))
            self._scalar_check = True

        elif plugin == IMAGE:
            self._image_writer.add((tag, step, value.image.encoded_image))
            self._image_writer.write()
            self._image_check = True



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Debugger train job register."""

import os

from mindinsight.datavisual.common.log import logger
from mindinsight.debugger.common.utils import is_valid_rank_dir_name
from mindinsight.utils.folder_analyzer import FolderAnalyzer


class DebuggerFolderAnalyzer(FolderAnalyzer):
    """Debugger train job register."""
    def analyze(self, entry, summary_base_dir, relative_path):
        """Check dir by debugger register."""
        update_info = {}
        if entry.is_dir():
            sub_relative_path = os.path.join(relative_path, entry.name)
            entry_path = entry.path
            try:
                subdir_entries = os.scandir(entry_path)
            except PermissionError:
                logger.warning('Path of %s under summary base directory is not accessible.', entry.name)
                return update_info
            subdir_entries = [subdir_entry for subdir_entry in subdir_entries if not subdir_entry.is_symlink()]
            subdir_entries = sorted(subdir_entries, key=lambda x: x.stat().st_mtime)
            for subdir_entry in subdir_entries:
                if subdir_entry.is_dir() and is_valid_rank_dir_name(subdir_entry.name) and \
                        os.path.exists(os.path.join(subdir_entry.path, ".dump_metadata")):
                    update_info = {'dump_dir': sub_relative_path}
                    return update_info
        return update_info
